/* Prepare a NAN string for strtod/strtof/strtold.
   Copyright (C) 2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <stdbool.h>
#include <string.h>

/* Scratch buffer used to store the result of nan_prepare_string.  */
struct nan_prepare_string_scratch
{
  char buffer[40];
};

/* Convert INPUT to a string of the form "NAN(...)".  Use *SCRATCH for
   storage.  Return an alternative string if the result would be too
   large for the scratch buffer.  */
static const char *
nan_prepare_string (const char *input,
                    struct nan_prepare_string_scratch *scratch)
{
  char *output = scratch->buffer;
  char *end = output + sizeof (scratch->buffer);

  /* The "NAN(" prefix.  */
  output[0] = 'N';
  output[1] = 'A';
  output[2] = 'N';
  output[3] = '(';
  output += 4;

  if (input[0] == '0')
    {
      /* Copy the radix prefix.  */
      output[0] = '0';
      bool hex;
      if (input[1] == 'x' || input[1] == 'X')
        {
          hex = true;
          output[1] = input[1];
          output += 2;
          input += 2;
        }
      else
        {
          hex = false;
          ++output;
          ++input;
        }

      /* Skip over all leading zeros, but leave one zero to preserve
         syntactic pecularities (such as "00x....").  */
      if (input[0] == '0')
        {
          output[0] = '0';
          ++output;
          do
            ++input;
          while (input[0] == '0');
        }

      /* Preserve invalid digits.  */
      if (hex)
        for (const char *p = input; *p != '\0'; ++p)
          {
            if (!(('0' <= *p && *p <= '9')
                  || ('a' <= *p && *p <= 'f')
                  || ('A' <= *p && *p <= 'F')))
              {
                output[0] = *p;
                output[1] = ')';
                output[2] = '\0';
                return scratch->buffer;
              }
          }
      else
        for (const char *p = input; *p != '\0'; ++p)
          {
            if (!('0' <= *p && *p <= '7'))
              {
                output[0] = *p;
                output[1] = ')';
                output[2] = '\0';
                return scratch->buffer;
              }
          }
    }
  else
    /* No leading zero.  Preserve non-decimal input.  */
    for (const char *p = input; *p != '\0'; ++p)
      {
        if (!('0' <= *p && *p <= '9'))
          {
            output[0] = *p;
            output[1] = ')';
            output[2] = '\0';
            return scratch->buffer;
          }
      }

  /* This string results in a NAN which has the same bit pattern as
     any other syntactically valid over-long NAN string, after
     conversion.  */
  const char *alternative = "NAN(0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF)";

  size_t length = strlen (input);
  /* Add space for the closing paren and the null terminator.  */
  if (length + 2 > end - output)
    return alternative;
  memcpy (output, input, length);
  output += length;
  output[0] = ')';
  output[1] = '\0';
  return scratch->buffer;
}
